---
title: Static-site generators
description: A guide to configuring popular static-site generators (Next.js, Hugo, VuePress, Middleman, Jekyll) for optimal deployment to IPFS, including best practices and configuration tips
---

# Static-site generators

Static-site generators like Hugo, Jekyll, Middleman, Next.js, and VuePress are all popular platforms for building static sites quickly. This guide walks through how to configure each of these generators to build for deployment to IPFS.

Check out the [IPFS Deploy GitHub Action Guide](./deploy-github-action.md) to automate the deployment of your static site to IPFS using GitHub Actions.

## Next.js

When deploying a Next.js site to IPFS, make sure that your site uses [Static Site Generation (SSG)](https://nextjs.org/docs/pages/building-your-application/rendering/static-site-generation), so that it can be [built as a static site](https://nextjs.org/docs/pages/building-your-application/deploying/static-exports).

1. First, ensure your `next.config.js` file has the following settings:

```js
module.exports = {
  output: 'export', // Enables static exports
  trailingSlash: true // Required for IPFS gateway compatibility
}
```

Key points about the configuration:

- `output: "export"` tells Next.js to generate a static site
- `trailingSlash: true` ensures routing works when the site is served from IPFS gateways (which require an index.html file per route)

To build your Next.js site:

```bash
npx next build
```

The static site will be generated in the `./out` directory.

### Important Considerations

- Only use [Static Site Generation (SSG)](https://nextjs.org/docs/pages/building-your-application/rendering/static-site-generation) features.
- Server-side features like `getServerSideProps` or API routes won't work.
- Dynamic routes need to be pre-rendered at build time.
- Use relative URLs for all internal links.

## Hugo

Refer to Hugo's [Quick Start](https://gohugo.io/getting-started/quick-start/) to install and set up your project.

In `config.toml` add `relativeURLs` and set it to `true`.

```
relativeURLs=true
```

Build static pages

```bash
hugo -D
```

Output will be in `./public/` directory by default. Upload the public folder to IPFS.

## VuePress

Refer to VuePress' [Getting Started](https://vuepress.vuejs.org/guide/getting-started.html) to install and set up your project.

To build a static site:

```bash
vuepress build
```

Output will be in `./.vuepress/dist` directory by default.

Use a command to convert a static site to only use relative URLs. In this example, we'll be using [all-relative](https://www.npmjs.com/package/all-relative)

```bash
cd .vuepress/dist/
npx all-relative
```

Upload the `dist` folder to IPFS.

## Middleman

Refer the Middleman's [Installation](https://middlemanapp.com/basics/install/) guide to install Ruby and Middleman.

1. Enable relative links and disable index file strip in your project's `config.rb` file:

   ```ruby
   set :relative_links, true
   set :strip_index_file, false
   ```

Links generated by the `link_to` helper or by Markdown will become relative.

1. Build your static site:

   ```bash
   middleman build
   ```

   Middleman will output your site to the `./build` folder.

1. Upload the `build` folder to IPFS.

## Jekyll

Refer to Jekyll's [Installation](https://jekyllrb.com/docs/installation/) guide to install Ruby and Jekyll.

Refer to Jekyll's [Quickstart](https://jekyllrb.com/docs/) to set up your project.

To build a static site:

```bash
jekyll build
```

Output will be in `./_site` by default.

Use a command to convert a static site to only use relative URLs. In this example, we'll be using [all-relative](https://www.npmjs.com/package/all-relative)

```bash
cd _site/
npx all-relative
```

Upload the `_site` folder to IPFS.

## WordPress

While WordPress is not a static site generator, it is possible to turn it into a static website allowing deployment to IPFS.

There are several plugins available to help you generate a static version of your WordPress site:

- [WP2Static](https://github.com/elementor/wp2static)
- [Simply Static](https://wordpress.org/plugins/simply-static/)
