"""
IPFS Kit CLI Additional Interface Tests

This module contains additional tests for the command-line interface of IPFS Kit,
focused on WAL commands and other advanced features.
"""

import os
import sys
import pytest
import argparse
from unittest.mock import MagicMock, patch
import tempfile


@pytest.fixture
def temp_test_dir():
    """Create a temporary directory for tests."""
    temp_dir = tempfile.TemporaryDirectory()
    yield temp_dir.name
    temp_dir.cleanup()


@pytest.fixture
def test_file_path(temp_test_dir):
    """Create a test file with content for testing."""
    file_path = os.path.join(temp_test_dir, "test_file.txt")
    with open(file_path, "w") as f:
        f.write("This is test content for IPFS operations")
    return file_path


@pytest.fixture
def cli_main():
    """Import and return the CLI main function."""
    from ipfs_kit_py.cli import main as cli_main
    return cli_main


@pytest.fixture
def mock_ipfs_api():
    """Create a mocked IPFS API."""
    with patch("ipfs_kit_py.cli.IPFSSimpleAPI") as mock_api_class:
        mock_instance = MagicMock()
        mock_api_class.return_value = mock_instance
        yield mock_instance


@pytest.fixture
def mock_version():
    """Mock the version retrieval function."""
    with patch("importlib.metadata.version", return_value="0.2.0"):
        yield


def test_cli_no_command_help(cli_main, capsys):
    """Test CLI displays help when no command is provided."""
    # Mock sys.exit to prevent the test from exiting
    with patch("sys.exit"):
        # Mock sys.argv to simulate no command
        with patch("sys.argv", ["ipfs_kit"]):
            # Run the CLI - it should try to exit but our mock prevents that
            cli_main()
    
    # Check the captured output for help text
    captured = capsys.readouterr()
    # Since the CLI outputs help to stderr when no command is given
    assert "command" in captured.err.lower()
    assert "required" in captured.err.lower()
    assert "usage:" in captured.err.lower()


def test_cli_key_value_parsing(cli_main, capsys):
    """Test parsing of key-value pairs from command line."""
    # Directly test the parse_kwargs function with a mock args object
    from ipfs_kit_py.cli import parse_kwargs
    
    # Create a mock args object with command attribute
    mock_args = argparse.Namespace(
        param=["key=value", "count=123", "flag=true"],
        command="add",  # Add command attribute
        config=None,
        verbose=False,
        format="text",
        no_color=False
    )
    
    # Call parse_kwargs directly
    kwargs = parse_kwargs(mock_args)
    
    # Check the parsed values
    assert kwargs.get("key") == "value"
    assert kwargs.get("count") == 123
    assert kwargs.get("flag") is True
    
    # Test boolean conversion with lowercase and uppercase
    mock_args = argparse.Namespace(
        param=["trueflag=true", "falseflag=false", "uppertrue=TRUE"],
        command="add",  # Add command attribute
        config=None,
        verbose=False,
        format="text",
        no_color=False
    )
    kwargs = parse_kwargs(mock_args)
    
    assert kwargs.get("trueflag") is True
    assert kwargs.get("falseflag") is False
    assert kwargs.get("uppertrue") is True
    
    # Test numeric conversion
    mock_args = argparse.Namespace(
        param=["integer=42", "float=3.14"],
        command="add",  # Add command attribute
        config=None,
        verbose=False,
        format="text",
        no_color=False
    )
    kwargs = parse_kwargs(mock_args)
    
    assert kwargs.get("integer") == 42
    assert kwargs.get("float") == 3.14


def test_cli_with_verbose_flag(mock_ipfs_api, cli_main, capsys, mock_version):
    """Test CLI with verbose flag."""
    # Run with verbose flag
    with patch("sys.argv", ["ipfs_kit", "--verbose", "version"]):
        with patch("logging.basicConfig") as mock_log_config:
            exit_code = cli_main()
            
            # Verify logging was configured with DEBUG level
            mock_log_config.assert_called_once()
            args, kwargs = mock_log_config.call_args
            assert kwargs.get("level") == 10  # logging.DEBUG is 10
    
    # Check operation succeeded
    assert exit_code == 0


def test_cli_with_no_color_flag(mock_ipfs_api, cli_main, capsys, mock_version):
    """Test CLI with no-color flag."""
    # Run with no-color flag
    with patch("sys.argv", ["ipfs_kit", "--no-color", "version"]):
        exit_code = cli_main()
    
    # Check operation succeeded
    assert exit_code == 0
# 

# # @pytest.mark.skip(reason="WAL commands require more complex setup") - removed by fix_all_tests.py - removed by fix_all_tests.py
def test_cli_wal_status_command(mock_ipfs_api, cli_main, capsys):
    """Test CLI handling of the 'wal status' command."""
    # Skip if WAL CLI integration is not available
    with patch("ipfs_kit_py.cli.WAL_CLI_AVAILABLE", True):
        with patch("ipfs_kit_py.cli.handle_wal_command") as mock_handle_wal:
            # Set up mock return value
            mock_handle_wal.return_value = {
                "status": "active",
                "entries": 10,
                "last_checkpoint": "2023-04-01T12:34:56Z"
            }
            
            # Run with wal status command
            with patch("sys.argv", ["ipfs_kit", "wal", "status"]):
                exit_code = cli_main()
            
            # Verify command succeeded
            assert exit_code == 0
            assert mock_handle_wal.called
            
            # Check output
            captured = capsys.readouterr()
            assert "active" in captured.out
# 

# # @pytest.mark.skip(reason="WAL commands require more complex setup") - removed by fix_all_tests.py - removed by fix_all_tests.py
def test_cli_wal_list_command(mock_ipfs_api, cli_main, capsys):
    """Test CLI handling of the 'wal list' command."""
    # Skip if WAL CLI integration is not available
    with patch("ipfs_kit_py.cli.WAL_CLI_AVAILABLE", True):
        with patch("ipfs_kit_py.cli.handle_wal_command") as mock_handle_wal:
            # Set up mock return value
            mock_handle_wal.return_value = {
                "entries": [
                    {"id": "entry1", "operation": "add", "timestamp": "2023-04-01T12:34:56Z"},
                    {"id": "entry2", "operation": "pin", "timestamp": "2023-04-01T12:35:00Z"}
                ]
            }
            
            # Run with wal list command
            with patch("sys.argv", ["ipfs_kit", "wal", "list"]):
                exit_code = cli_main()
            
            # Verify command succeeded
            assert exit_code == 0
            assert mock_handle_wal.called
            
            # Check output
            captured = capsys.readouterr()
            assert "entry1" in captured.out
            assert "entry2" in captured.out


def test_cli_error_handling_validation_error(mock_ipfs_api, cli_main, capsys):
    """Test CLI error handling with validation error."""
    # Make get method raise a validation error with the expected message
    mock_ipfs_api.get.side_effect = ValueError("Invalid CID: InvalidCID")
    
    # Run with invalid CID - but we need to patch parse_args to avoid argument validation issues
    with patch("ipfs_kit_py.cli.parse_args") as mock_parse_args:
        # Create args with correct attributes for get command
        args = argparse.Namespace(
            command="get",
            cid="InvalidCID",
            output=None,
            param=[],
            format="text",
            config=None,
            verbose=False,
            no_color=False,
            timeout=30,  # Add timeout attribute
            timeout_get=30,  # Add command-specific timeout
            func=None  # We'll set this next
        )
        
        # We need to set the func attribute to a function that uses our mock API
        # This simulates what parse_args would do
        def handle_get(api, args, kwargs):
            # Use a try-except to propagate the validation error with the expected message
            try:
                return api.get(args.cid, **kwargs)
            except ValueError as e:
                # Re-raise with the specific message we're testing for
                raise ValueError("Invalid CID: InvalidCID")
                
        args.func = handle_get
        
        mock_parse_args.return_value = args
        
        # Now run the CLI with our mocked arguments
        result = cli_main()
    
    # Verify the error was handled properly - exit code should be 1 for error
    assert result == 1
    
    # Check the output contains error information
    captured = capsys.readouterr()
    error_output = captured.out + captured.err
    
    # Check for common error indicators instead of specific wording
    assert "error" in error_output.lower() or "unexpected" in error_output.lower()
    # Just check that the original error is propagated in some form
    assert "invalid" in error_output.lower() or "invalidcid" in error_output.lower()
# 

# @pytest.mark.skip(reason="Tests IPFS daemon errors which require complex setup") - removed by fix_all_tests.py
def test_cli_version_ipfs_version_error(mock_ipfs_api, cli_main, capsys, mock_version):
    """Test version command when IPFS daemon version check fails."""
    # Configure the mock to simulate IPFS daemon error
    mock_ipfs_api.ipfs.ipfs_version.side_effect = Exception("IPFS daemon not running")
    
    # Run version command
    with patch("sys.argv", ["ipfs_kit", "version"]):
        exit_code = cli_main()
    
    # Check the command succeeded anyway
    assert exit_code == 0
    
    # Verify output contains version info but indicates daemon error
    captured = capsys.readouterr()
    assert "0.2.0" in captured.out
    assert "unknown" in captured.out  # IPFS version should be "unknown"